#   BSD LICENSE
#
#   Copyright(c) 2010-2014 Intel Corporation. All rights reserved.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     * Neither the name of Intel Corporation nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The main logic behind running autotests in parallel

from __future__ import print_function
import StringIO
import csv
import multiprocessing
import pexpect
import re
import subprocess
import sys
import time

# wait for prompt


def wait_prompt(child):
    try:
        child.sendline()
        result = child.expect(["RTE>>", pexpect.TIMEOUT, pexpect.EOF],
                              timeout=120)
    except:
        return False
    if result == 0:
        return True
    else:
        return False

# run a test group
# each result tuple in results list consists of:
#   result value (0 or -1)
#   result string
#   test name
#   total test run time (double)
#   raw test log
#   test report (if not available, should be None)
#
# this function needs to be outside AutotestRunner class
# because otherwise Pool won't work (or rather it will require
# quite a bit of effort to make it work).


def run_test_group(cmdline, target, test_group):
    results = []
    child = None
    start_time = time.time()
    startuplog = None

    # run test app
    try:
        # prepare logging of init
        startuplog = StringIO.StringIO()

        print("\n%s %s\n" % ("=" * 20, test_group["Prefix"]), file=startuplog)
        print("\ncmdline=%s" % cmdline, file=startuplog)

        child = pexpect.spawn(cmdline, logfile=startuplog)

        # wait for target to boot
        if not wait_prompt(child):
            child.close()

            results.append((-1,
                            "Fail [No prompt]",
                            "Start %s" % test_group["Prefix"],
                            time.time() - start_time,
                            startuplog.getvalue(),
                            None))

            # mark all tests as failed
            for test in test_group["Tests"]:
                results.append((-1, "Fail [No prompt]", test["Name"],
                                time.time() - start_time, "", None))
            # exit test
            return results

    except:
        results.append((-1,
                        "Fail [Can't run]",
                        "Start %s" % test_group["Prefix"],
                        time.time() - start_time,
                        startuplog.getvalue(),
                        None))

        # mark all tests as failed
        for t in test_group["Tests"]:
            results.append((-1, "Fail [Can't run]", t["Name"],
                            time.time() - start_time, "", None))
        # exit test
        return results

    # startup was successful
    results.append((0, "Success", "Start %s" % test_group["Prefix"],
                    time.time() - start_time, startuplog.getvalue(), None))

    # run all tests in test group
    for test in test_group["Tests"]:

        # create log buffer for each test
        # in multiprocessing environment, the logging would be
        # interleaved and will create a mess, hence the buffering
        logfile = StringIO.StringIO()
        child.logfile = logfile

        result = ()

        # make a note when the test started
        start_time = time.time()

        try:
            # print test name to log buffer
            print("\n%s %s\n" % ("-" * 20, test["Name"]), file=logfile)

            # run test function associated with the test
            result = test["Func"](child, test["Command"])

            # make a note when the test was finished
            end_time = time.time()

            log = logfile.getvalue()

            # append test data to the result tuple
            result += (test["Name"], end_time - start_time, log)

            # call report function, if any defined, and supply it with
            # target and complete log for test run
            if test["Report"]:
                report = test["Report"](target, log)

                # append report to results tuple
                result += (report,)
            else:
                # report is None
                result += (None,)
        except:
            # make a note when the test crashed
            end_time = time.time()

            # mark test as failed
            result = (-1, "Fail [Crash]", test["Name"],
                      end_time - start_time, logfile.getvalue(), None)
        finally:
            # append the results to the results list
            results.append(result)

    # regardless of whether test has crashed, try quitting it
    try:
        child.sendline("quit")
        child.close()
    # if the test crashed, just do nothing instead
    except:
        # nop
        pass

    # return test results
    return results


# class representing an instance of autotests run
class AutotestRunner:
    cmdline = ""
    parallel_test_groups = []
    non_parallel_test_groups = []
    logfile = None
    csvwriter = None
    target = ""
    start = None
    n_tests = 0
    fails = 0
    log_buffers = []
    blacklist = []
    whitelist = []

    def __init__(self, cmdline, target, blacklist, whitelist):
        self.cmdline = cmdline
        self.target = target
        self.binary = cmdline.split()[0]
        self.blacklist = blacklist
        self.whitelist = whitelist
        self.skipped = []

        # log file filename
        logfile = "%s.log" % target
        csvfile = "%s.csv" % target

        self.logfile = open(logfile, "w")
        csvfile = open(csvfile, "w")
        self.csvwriter = csv.writer(csvfile)

        # prepare results table
        self.csvwriter.writerow(["test_name", "test_result", "result_str"])

    # set up cmdline string
    def __get_cmdline(self, test):
        cmdline = self.cmdline

        # append memory limitations for each test
        # otherwise tests won't run in parallel
        if "i686" not in self.target:
            cmdline += " --socket-mem=%s" % test["Memory"]
        else:
            # affinitize startup so that tests don't fail on i686
            cmdline = "taskset 1 " + cmdline
            cmdline += " -m " + str(sum(map(int, test["Memory"].split(","))))

        # set group prefix for autotest group
        # otherwise they won't run in parallel
        cmdline += " --file-prefix=%s" % test["Prefix"]

        return cmdline

    def add_parallel_test_group(self, test_group):
        self.parallel_test_groups.append(test_group)

    def add_non_parallel_test_group(self, test_group):
        self.non_parallel_test_groups.append(test_group)

    def __process_results(self, results):
        # this iterates over individual test results
        for i, result in enumerate(results):

            # increase total number of tests that were run
            # do not include "start" test
            if i > 0:
                self.n_tests += 1

            # unpack result tuple
            test_result, result_str, test_name, \
                test_time, log, report = result

            # get total run time
            cur_time = time.time()
            total_time = int(cur_time - self.start)

            # print results, test run time and total time since start
            result = ("%s:" % test_name).ljust(30)
            result += result_str.ljust(29)
            result += "[%02dm %02ds]" % (test_time / 60, test_time % 60)

            # don't print out total time every line, it's the same anyway
            if i == len(results) - 1:
                print(result +
                      "[%02dm %02ds]" % (total_time / 60, total_time % 60))
            else:
                print(result)

            # if test failed and it wasn't a "start" test
            if test_result < 0 and not i == 0:
                self.fails += 1

            # collect logs
            self.log_buffers.append(log)

            # create report if it exists
            if report:
                try:
                    f = open("%s_%s_report.rst" %
                             (self.target, test_name), "w")
                except IOError:
                    print("Report for %s could not be created!" % test_name)
                else:
                    with f:
                        f.write(report)

            # write test result to CSV file
            if i != 0:
                self.csvwriter.writerow([test_name, test_result, result_str])

    # this function checks individual test and decides if this test should be in
    # the group by comparing it against  whitelist/blacklist. it also checks if
    # the test is compiled into the binary, and marks it as skipped if necessary
    def __filter_test(self, test):
        test_cmd = test["Command"]
        test_id = test_cmd

        # dump tests are specified in full e.g. "Dump_mempool"
        if "_autotest" in test_id:
            test_id = test_id[:-len("_autotest")]

        # filter out blacklisted/whitelisted tests
        if self.blacklist and test_id in self.blacklist:
            return False
        if self.whitelist and test_id not in self.whitelist:
            return False

        # if test wasn't compiled in, remove it as well

        # parse the binary for available test commands
        stripped = 'not stripped' not in \
                   subprocess.check_output(['file', self.binary])
        if not stripped:
            symbols = subprocess.check_output(['nm',
                                               self.binary]).decode('utf-8')
            avail_cmds = re.findall('test_register_(\w+)', symbols)

            if test_cmd not in avail_cmds:
                # notify user
                result = 0, "Skipped [Not compiled]", test_id, 0, "", None
                self.skipped.append(tuple(result))
                return False

        return True

    def __filter_group(self, group):
        group["Tests"] = list(filter(self.__filter_test, group["Tests"]))
        return len(group["Tests"]) > 0

    # iterate over test groups and run tests associated with them
    def run_all_tests(self):
        # filter groups
        # for each test group, check all tests against the filter, then remove
        # all groups that don't have any tests
        self.parallel_test_groups = list(
            filter(self.__filter_group,
                   self.parallel_test_groups)
        )
        self.non_parallel_test_groups = list(
            filter(self.__filter_group,
                   self.non_parallel_test_groups)
        )

        # create a pool of worker threads
        pool = multiprocessing.Pool(processes=1)

        results = []

        # whatever happens, try to save as much logs as possible
        try:

            # create table header
            print("")
            print("Test name".ljust(30) + "Test result".ljust(29) +
                  "Test".center(9) + "Total".center(9))
            print("=" * 80)

            # print out skipped autotests if there were any
            if len(self.skipped):
                print("Skipped autotests:")

                # print out any skipped tests
                for result in self.skipped:
                    # unpack result tuple
                    test_result, result_str, test_name, _, _, _ = result
                    self.csvwriter.writerow([test_name, test_result,
                                             result_str])

                    t = ("%s:" % test_name).ljust(30)
                    t += result_str.ljust(29)
                    t += "[00m 00s]"

                    print(t)

            # make a note of tests start time
            self.start = time.time()

            print("Parallel autotests:")
            # assign worker threads to run test groups
            for test_group in self.parallel_test_groups:
                result = pool.apply_async(run_test_group,
                                          [self.__get_cmdline(test_group),
                                           self.target,
                                           test_group])
                results.append(result)

            # iterate while we have group execution results to get
            while len(results) > 0:

                # iterate over a copy to be able to safely delete results
                # this iterates over a list of group results
                for group_result in results[:]:

                    # if the thread hasn't finished yet, continue
                    if not group_result.ready():
                        continue

                    res = group_result.get()

                    self.__process_results(res)

                    # remove result from results list once we're done with it
                    results.remove(group_result)

            print("Non-parallel autotests:")
            # run non_parallel tests. they are run one by one, synchronously
            for test_group in self.non_parallel_test_groups:
                group_result = run_test_group(
                    self.__get_cmdline(test_group), self.target, test_group)

                self.__process_results(group_result)

            # get total run time
            cur_time = time.time()
            total_time = int(cur_time - self.start)

            # print out summary
            print("=" * 80)
            print("Total run time: %02dm %02ds" % (total_time / 60,
                                                   total_time % 60))
            if self.fails != 0:
                print("Number of failed tests: %s" % str(self.fails))

            # write summary to logfile
            self.logfile.write("Summary\n")
            self.logfile.write("Target: ".ljust(15) + "%s\n" % self.target)
            self.logfile.write("Tests: ".ljust(15) + "%i\n" % self.n_tests)
            self.logfile.write("Failed tests: ".ljust(
                15) + "%i\n" % self.fails)
        except:
            print("Exception occurred")
            print(sys.exc_info())
            self.fails = 1

        # drop logs from all executions to a logfile
        for buf in self.log_buffers:
            self.logfile.write(buf.replace("\r", ""))

        return self.fails



#   BSD LICENSE
#
#   Copyright(c) 2010-2014 Intel Corporation. All rights reserved.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     * Neither the name of Intel Corporation nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Test data for autotests

from glob import glob
from autotest_test_funcs import *


# quick and dirty function to find out number of sockets
def num_sockets():
    result = len(glob("/sys/devices/system/node/node*"))
    if result == 0:
        return 1
    return result


# Assign given number to each socket
# e.g. 32 becomes 32,32 or 32,32,32,32
def per_sockets(num):
    return ",".join([str(num)] * num_sockets())

# groups of tests that can be run in parallel
# the grouping has been found largely empirically
parallel_test_group_list = [
    {
        "Prefix":    "group_1",
        "Memory":    per_sockets(8),
        "Tests":
        [
            {
                "Name":    "Cycles autotest",
                "Command": "cycles_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Timer autotest",
                "Command": "timer_autotest",
                "Func":    timer_autotest,
                "Report":   None,
            },
            {
                "Name":    "Debug autotest",
                "Command": "debug_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Errno autotest",
                "Command": "errno_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Meter autotest",
                "Command": "meter_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Common autotest",
                "Command": "common_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Resource autotest",
                "Command": "resource_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "group_2",
        "Memory":    "16",
        "Tests":
        [
            {
                "Name":    "Memory autotest",
                "Command": "memory_autotest",
                "Func":    memory_autotest,
                "Report":  None,
            },
            {
                "Name":    "Read/write lock autotest",
                "Command": "rwlock_autotest",
                "Func":    rwlock_autotest,
                "Report":  None,
            },
            {
                "Name":    "Logs autotest",
                "Command": "logs_autotest",
                "Func":    logs_autotest,
                "Report":  None,
            },
            {
                "Name":    "CPU flags autotest",
                "Command": "cpuflags_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Version autotest",
                "Command": "version_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "EAL filesystem autotest",
                "Command": "eal_fs_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "EAL flags autotest",
                "Command": "eal_flags_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Hash autotest",
                "Command": "hash_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ],
    },
    {
        "Prefix":    "group_3",
        "Memory":    per_sockets(512),
        "Tests":
        [
            {
                "Name":    "LPM autotest",
                "Command": "lpm_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "LPM6 autotest",
                "Command": "lpm6_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Memcpy autotest",
                "Command": "memcpy_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Memzone autotest",
                "Command": "memzone_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "String autotest",
                "Command": "string_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Alarm autotest",
                "Command": "alarm_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "group_4",
        "Memory":    per_sockets(128),
        "Tests":
        [
            {
                "Name":    "PCI autotest",
                "Command": "pci_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Malloc autotest",
                "Command": "malloc_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Multi-process autotest",
                "Command": "multiprocess_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Mbuf autotest",
                "Command": "mbuf_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Per-lcore autotest",
                "Command": "per_lcore_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Ring autotest",
                "Command": "ring_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "group_5",
        "Memory":    "32",
        "Tests":
        [
            {
                "Name":    "Spinlock autotest",
                "Command": "spinlock_autotest",
                "Func":    spinlock_autotest,
                "Report":  None,
            },
            {
                "Name":    "Byte order autotest",
                "Command": "byteorder_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "TAILQ autotest",
                "Command": "tailq_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Command-line autotest",
                "Command": "cmdline_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Interrupts autotest",
                "Command": "interrupt_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "group_6",
        "Memory":    per_sockets(512),
        "Tests":
        [
            {
                "Name":    "Function reentrancy autotest",
                "Command": "func_reentrancy_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Mempool autotest",
                "Command": "mempool_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Atomics autotest",
                "Command": "atomic_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Prefetch autotest",
                "Command": "prefetch_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Red autotest",
                "Command": "red_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "group_7",
        "Memory":    "64",
        "Tests":
        [
            {
                "Name":    "PMD ring autotest",
                "Command": "ring_pmd_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Access list control autotest",
                "Command": "acl_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
            {
                "Name":    "Sched autotest",
                "Command": "sched_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
]

# tests that should not be run when any other tests are running
non_parallel_test_group_list = [

    {
        "Prefix":    "eventdev",
        "Memory":    "512",
        "Tests":
        [
            {
                "Name":    "Eventdev common autotest",
                "Command": "eventdev_common_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "eventdev_sw",
        "Memory":    "512",
        "Tests":
        [
            {
                "Name":    "Eventdev sw autotest",
                "Command": "eventdev_sw_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "kni",
        "Memory":    "512",
        "Tests":
        [
            {
                "Name":    "KNI autotest",
                "Command": "kni_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "mempool_perf",
        "Memory":    per_sockets(256),
        "Tests":
        [
            {
                "Name":    "Mempool performance autotest",
                "Command": "mempool_perf_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "memcpy_perf",
        "Memory":    per_sockets(512),
        "Tests":
        [
            {
                "Name":    "Memcpy performance autotest",
                "Command": "memcpy_perf_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":    "hash_perf",
        "Memory":    per_sockets(512),
        "Tests":
        [
            {
                "Name":    "Hash performance autotest",
                "Command": "hash_perf_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
    {
        "Prefix":      "power",
        "Memory":      "16",
        "Tests":
        [
            {
                "Name":       "Power autotest",
                "Command":    "power_autotest",
                "Func":       default_autotest,
                "Report":      None,
            },
        ]
    },
    {
        "Prefix":      "power_acpi_cpufreq",
        "Memory":      "16",
        "Tests":
        [
            {
                "Name":       "Power ACPI cpufreq autotest",
                "Command":    "power_acpi_cpufreq_autotest",
                "Func":       default_autotest,
                "Report":     None,
            },
        ]
    },
    {
        "Prefix":      "power_kvm_vm",
        "Memory":      "16",
        "Tests":
        [
            {
                "Name":       "Power KVM VM  autotest",
                "Command":    "power_kvm_vm_autotest",
                "Func":       default_autotest,
                "Report":     None,
            },
        ]
    },
    {
        "Prefix":    "timer_perf",
        "Memory":    per_sockets(512),
        "Tests":
        [
            {
                "Name":    "Timer performance autotest",
                "Command": "timer_perf_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },

    #
    # Please always make sure that ring_perf is the last test!
    #
    {
        "Prefix":    "ring_perf",
        "Memory":    per_sockets(512),
        "Tests":
        [
            {
                "Name":    "Ring performance autotest",
                "Command": "ring_perf_autotest",
                "Func":    default_autotest,
                "Report":  None,
            },
        ]
    },
]



#   BSD LICENSE
#   Copyright(c) 2010-2015 Intel Corporation. All rights reserved.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#   * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#   * Neither the name of Intel Corporation nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
import subprocess
from docutils import nodes
from distutils.version import LooseVersion
from sphinx import __version__ as sphinx_version
from sphinx.highlighting import PygmentsBridge
from pygments.formatters.latex import LatexFormatter
from os import listdir
from os import environ
from os.path import basename
from os.path import dirname
from os.path import join as path_join

try:
    # Python 2.
    import ConfigParser as configparser
except:
    # Python 3.
    import configparser

try:
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
except:
    print('Install the sphinx ReadTheDocs theme for improved html documentation '
          'layout: pip install sphinx_rtd_theme')
    pass

project = 'Data Plane Development Kit'
html_logo = '../logo/DPDK_logo_vertical_rev_small.png'
latex_logo = '../logo/DPDK_logo_horizontal_tag.png'
html_add_permalinks = ""
html_show_copyright = False
highlight_language = 'none'

# If MAKEFLAGS is exported by the user, garbage text might end up in version
version = subprocess.check_output(['make', '-sRrC', '../../', 'showversion'],
                                  env=dict(environ, MAKEFLAGS=""))
version = version.decode('utf-8').rstrip()
release = version

master_doc = 'index'

# Maximum feature description string length
feature_str_len = 25

# Figures, tables and code-blocks automatically numbered if they have caption
numfig = True

latex_documents = [
    ('index',
     'doc.tex',
     '',
     '',
     'manual')
]

# Latex directives to be included directly in the latex/pdf docs.
custom_latex_preamble = r"""
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\usepackage{helvet}
\renewcommand{\familydefault}{\sfdefault}
\RecustomVerbatimEnvironment{Verbatim}{Verbatim}{xleftmargin=5mm}
"""

# Configuration for the latex/pdf docs.
latex_elements = {
    'papersize': 'a4paper',
    'pointsize': '11pt',
    # remove blank pages
    'classoptions': ',openany,oneside',
    'babel': '\\usepackage[english]{babel}',
    # customize Latex formatting
    'preamble': custom_latex_preamble
}


# Override the default Latex formatter in order to modify the
# code/verbatim blocks.
class CustomLatexFormatter(LatexFormatter):
    def __init__(self, **options):
        super(CustomLatexFormatter, self).__init__(**options)
        # Use the second smallest font size for code/verbatim blocks.
        self.verboptions = r'formatcom=\footnotesize'

# Replace the default latex formatter.
PygmentsBridge.latex_formatter = CustomLatexFormatter

# Configuration for man pages
man_pages = [("testpmd_app_ug/run_app", "testpmd",
              "tests for dpdk pmds", "", 1),
             ("tools/pdump", "dpdk-pdump",
              "enable packet capture on dpdk ports", "", 1),
             ("tools/proc_info", "dpdk-procinfo",
              "access dpdk port stats and memory info", "", 1),
             ("tools/pmdinfo", "dpdk-pmdinfo",
              "dump a PMDs hardware support info", "", 1),
             ("tools/devbind", "dpdk-devbind",
              "check device status and bind/unbind them from drivers", "", 8)]


# ####### :numref: fallback ########
# The following hook functions add some simple handling for the :numref:
# directive for Sphinx versions prior to 1.3.1. The functions replace the
# :numref: reference with a link to the target (for all Sphinx doc types).
# It doesn't try to label figures/tables.
def numref_role(reftype, rawtext, text, lineno, inliner):
    """
    Add a Sphinx role to handle numref references. Note, we can't convert
    the link here because the doctree isn't build and the target information
    isn't available.
    """
    # Add an identifier to distinguish numref from other references.
    newnode = nodes.reference('',
                              '',
                              refuri='_local_numref_#%s' % text,
                              internal=True)
    return [newnode], []


def process_numref(app, doctree, from_docname):
    """
    Process the numref nodes once the doctree has been built and prior to
    writing the files. The processing involves replacing the numref with a
    link plus text to indicate if it is a Figure or Table link.
    """

    # Iterate over the reference nodes in the doctree.
    for node in doctree.traverse(nodes.reference):
        target = node.get('refuri', '')

        # Look for numref nodes.
        if target.startswith('_local_numref_#'):
            target = target.replace('_local_numref_#', '')

            # Get the target label and link information from the Sphinx env.
            data = app.builder.env.domains['std'].data
            docname, label, _ = data['labels'].get(target, ('', '', ''))
            relative_url = app.builder.get_relative_uri(from_docname, docname)

            # Add a text label to the link.
            if target.startswith('figure'):
                caption = 'Figure'
            elif target.startswith('table'):
                caption = 'Table'
            else:
                caption = 'Link'

            # New reference node with the updated link information.
            newnode = nodes.reference('',
                                      caption,
                                      refuri='%s#%s' % (relative_url, label),
                                      internal=True)
            node.replace_self(newnode)


def generate_overview_table(output_filename, table_id, section, table_name, title):
    """
    Function to generate the Overview Table from the ini files that define
    the features for each driver.

    The default features for the table and their order is defined by the
    'default.ini' file.

    """
    # Default warning string.
    warning = 'Warning generate_overview_table()'

    # Get the default features and order from the 'default.ini' file.
    ini_path = path_join(dirname(output_filename), 'features')
    config = configparser.ConfigParser()
    config.optionxform = str
    config.read(path_join(ini_path, 'default.ini'))
    default_features = config.items(section)

    # Create a dict of the valid features to validate the other ini files.
    valid_features = {}
    max_feature_length = 0
    for feature in default_features:
        key = feature[0]
        valid_features[key] = ' '
        max_feature_length = max(max_feature_length, len(key))

    # Get a list of driver ini files, excluding 'default.ini'.
    ini_files = [basename(file) for file in listdir(ini_path)
                 if file.endswith('.ini') and file != 'default.ini']
    ini_files.sort()

    # Build up a list of the table header names from the ini filenames.
    header_names = []
    for ini_filename in ini_files:
        name = ini_filename[:-4]
        name = name.replace('_vf', 'vf')

        # Pad the table header names to match the existing format.
        if '_vec' in name:
            pmd, vec = name.split('_')
            name = '{0:{fill}{align}7}vec'.format(pmd, fill='.', align='<')
        else:
            name = '{0:{fill}{align}10}'.format(name, fill=' ', align='<')

        header_names.append(name)

    # Create a dict of the defined features for each driver from the ini files.
    ini_data = {}
    for ini_filename in ini_files:
        config = configparser.ConfigParser()
        config.optionxform = str
        config.read(path_join(ini_path, ini_filename))

        # Initialize the dict with the default.ini value.
        ini_data[ini_filename] = valid_features.copy()

        # Check for a valid ini section.
        if not config.has_section(section):
            print("{}: File '{}' has no [{}] secton".format(warning,
                                                            ini_filename,
                                                            section))
            continue

        # Check for valid features names.
        for name, value in config.items(section):
            if name not in valid_features:
                print("{}: Unknown feature '{}' in '{}'".format(warning,
                                                                name,
                                                                ini_filename))
                continue

            if value is not '':
                # Get the first letter only.
                ini_data[ini_filename][name] = value[0]

    # Print out the RST Driver Overview table from the ini file data.
    outfile = open(output_filename, 'w')
    num_cols = len(header_names)

    print_table_css(outfile, table_id)
    print('.. table:: ' + table_name + '\n', file=outfile)
    print_table_header(outfile, num_cols, header_names, title)
    print_table_body(outfile, num_cols, ini_files, ini_data, default_features)


def print_table_header(outfile, num_cols, header_names, title):
    """ Print the RST table header. The header names are vertical. """
    print_table_divider(outfile, num_cols)

    line = ''
    for name in header_names:
        line += ' ' + name[0]

    print_table_row(outfile, title, line)

    for i in range(1, 10):
        line = ''
        for name in header_names:
            line += ' ' + name[i]

        print_table_row(outfile, '', line)

    print_table_divider(outfile, num_cols)


def print_table_body(outfile, num_cols, ini_files, ini_data, default_features):
    """ Print out the body of the table. Each row is a NIC feature. """

    for feature, _ in default_features:
        line = ''

        for ini_filename in ini_files:
            line += ' ' + ini_data[ini_filename][feature]

        print_table_row(outfile, feature, line)

    print_table_divider(outfile, num_cols)


def print_table_row(outfile, feature, line):
    """ Print a single row of the table with fixed formatting. """
    line = line.rstrip()
    print('   {:<{}}{}'.format(feature, feature_str_len, line), file=outfile)


def print_table_divider(outfile, num_cols):
    """ Print the table divider line. """
    line = ' '
    column_dividers = ['='] * num_cols
    line += ' '.join(column_dividers)

    feature = '=' * feature_str_len

    print_table_row(outfile, feature, line)


def print_table_css(outfile, table_id):
    template = """
.. raw:: html

   <style>
      .wy-nav-content {
         opacity: .99;
      }
      table#idx {
         cursor: default;
         overflow: hidden;
      }
      table#idx th, table#idx td {
         text-align: center;
      }
      table#idx th {
         font-size: 80%;
         white-space: pre-wrap;
         vertical-align: top;
         padding: 0.5em 0;
         min-width: 0.9em;
         width: 2em;
      }
      table#idx col:first-child {
         width: 0;
      }
      table#idx th:first-child {
         vertical-align: bottom;
      }
      table#idx td {
         font-size: 70%;
         padding: 1px;
      }
      table#idx td:first-child {
         padding-left: 1em;
         text-align: left;
      }
      table#idx tr:nth-child(2n-1) td {
         background-color: rgba(210, 210, 210, 0.2);
      }
      table#idx th:not(:first-child):hover,
      table#idx td:not(:first-child):hover {
         position: relative;
      }
      table#idx th:not(:first-child):hover::after,
      table#idx td:not(:first-child):hover::after {
         content: '';
         height: 6000px;
         top: -3000px;
         width: 100%;
         left: 0;
         position: absolute;
         z-index: -1;
         background-color: #ffb;
      }
      table#idx tr:hover td {
         background-color: #ffb;
      }
   </style>
"""
    print(template.replace("idx", "id%d" % (table_id)), file=outfile)


def setup(app):
    table_file = dirname(__file__) + '/nics/overview_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in networking drivers',
                            'Feature')
    table_file = dirname(__file__) + '/cryptodevs/overview_feature_table.txt'
    generate_overview_table(table_file, 1,
                            'Features',
                            'Features availability in crypto drivers',
                            'Feature')
    table_file = dirname(__file__) + '/cryptodevs/overview_cipher_table.txt'
    generate_overview_table(table_file, 2,
                            'Cipher',
                            'Cipher algorithms in crypto drivers',
                            'Cipher algorithm')
    table_file = dirname(__file__) + '/cryptodevs/overview_auth_table.txt'
    generate_overview_table(table_file, 3,
                            'Auth',
                            'Authentication algorithms in crypto drivers',
                            'Authentication algorithm')
    table_file = dirname(__file__) + '/cryptodevs/overview_aead_table.txt'
    generate_overview_table(table_file, 4,
                            'AEAD',
                            'AEAD algorithms in crypto drivers',
                            'AEAD algorithm')

    if LooseVersion(sphinx_version) < LooseVersion('1.3.1'):
        print('Upgrade sphinx to version >= 1.3.1 for '
              'improved Figure/Table number handling.')
        # Add a role to handle :numref: references.
        app.add_role('numref', numref_role)
        # Process the numref references once the doctree has been created.
        app.connect('doctree-resolved', process_numref)



#!/usr/bin/env python

#   BSD LICENSE
#
#   Copyright(c) 2016 Intel Corporation. All rights reserved.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     * Neither the name of Intel Corporation nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# This script maps the set of pipelines identified (MASTER pipelines are
# ignored) from the input configuration file to the set of cores
# provided as input argument and creates configuration files for each of
# the mapping combinations.
#

from __future__ import print_function
from collections import namedtuple
import argparse
import array
import errno
import itertools
import os
import re
import sys

# default values
enable_stage0_traceout = 1
enable_stage1_traceout = 1
enable_stage2_traceout = 1

enable_stage1_fileout = 1
enable_stage2_fileout = 1

Constants = namedtuple('Constants', ['MAX_CORES', 'MAX_PIPELINES'])
constants = Constants(16, 64)

# pattern for physical core
pattern_phycore = '^(s|S)\d(c|C)[1-9][0-9]*$'
reg_phycore = re.compile(pattern_phycore)


def popcount(mask):
    return bin(mask).count("1")


def len2mask(length):
    if (length == 0):
        return 0

    if (length > 64):
        sys.exit('error: len2mask - length %i > 64. exiting' % length)

    return int('1' * length, 2)


def bitstring_write(n, n_bits):
    tmpstr = ""
    if (n_bits > 64):
        return

    i = n_bits - 1
    while (i >= 0):
        cond = (n & (1 << i))
        if (cond):
            print('1', end='')
            tmpstr += '1'
        else:
            print('0', end='')
            tmpstr += '0'
        i -= 1
    return tmpstr


class Cores0:

    def __init__(self):
        self.n_pipelines = 0


class Cores1:

    def __init__(self):
        self.pipelines = 0
        self.n_pipelines = 0


class Cores2:

    def __init__(self):
        self.pipelines = 0
        self.n_pipelines = 0
        self.counter = 0
        self.counter_max = 0
        self.bitpos = array.array(
            "L", itertools.repeat(0, constants.MAX_PIPELINES))


class Context0:

    def __init__(self):
        self.cores = [Cores0() for i in range(0, constants.MAX_CORES)]
        self.n_cores = 0
        self.n_pipelines = 0
        self.n_pipelines0 = 0
        self.pos = 0
        self.file_comment = ""
        self.ctx1 = None
        self.ctx2 = None

    def stage0_print(self):
        print('printing Context0 obj')
        print('c0.cores(n_pipelines) = [ ', end='')
        for cores_count in range(0, constants.MAX_CORES):
            print(self.cores[cores_count].n_pipelines, end=' ')
        print(']')
        print('c0.n_cores = %d' % self.n_cores)
        print('c0.n_pipelines = %d' % self.n_pipelines)
        print('c0.n_pipelines0 = %d' % self.n_pipelines0)
        print('c0.pos = %d' % self.pos)
        print('c0.file_comment = %s' % self.file_comment)
        if (self.ctx1 is not None):
            print('c0.ctx1 = ', end='')
            print(repr(self.ctx1))
        else:
            print('c0.ctx1 = None')

        if (self.ctx2 is not None):
            print('c0.ctx2 = ', end='')
            print(repr(self.ctx2))
        else:
            print('c0.ctx2 = None')

    def stage0_init(self, num_cores, num_pipelines, ctx1, ctx2):
        self.n_cores = num_cores
        self.n_pipelines = num_pipelines
        self.ctx1 = ctx1
        self.ctx2 = ctx2

    def stage0_process(self):
        # stage0 init
        self.cores[0].n_pipelines = self.n_pipelines
        self.n_pipelines0 = 0
        self.pos = 1

        while True:
            # go forward
            while True:
                if ((self.pos < self.n_cores) and (self.n_pipelines0 > 0)):
                    self.cores[self.pos].n_pipelines = min(
                        self.cores[self.pos - 1].n_pipelines,
                        self.n_pipelines0)
                    self.n_pipelines0 -= self.cores[self.pos].n_pipelines
                    self.pos += 1
                else:
                    break

            # check solution
            if (self.n_pipelines0 == 0):
                self.stage0_log()
                self.ctx1.stage1_init(self, self.ctx2)  # self is object c0
                self.ctx1.stage1_process()

            # go backward
            while True:
                if (self.pos == 0):
                    return

                self.pos -= 1
                if ((self.cores[self.pos].n_pipelines > 1) and
                        (self.pos != (self.n_cores - 1))):
                    break

                self.n_pipelines0 += self.cores[self.pos].n_pipelines
                self.cores[self.pos].n_pipelines = 0

            # rearm
            self.cores[self.pos].n_pipelines -= 1
            self.n_pipelines0 += 1
            self.pos += 1

    def stage0_log(self):
        tmp_file_comment = ""
        if(enable_stage0_traceout != 1):
            return

        print('STAGE0: ', end='')
        tmp_file_comment += 'STAGE0: '
        for cores_count in range(0, self.n_cores):
            print('C%d = %d\t'
                  % (cores_count,
                      self.cores[cores_count].n_pipelines), end='')
            tmp_file_comment += "C{} = {}\t".format(
                cores_count, self.cores[cores_count].n_pipelines)
        # end for
        print('')
        self.ctx1.stage0_file_comment = tmp_file_comment
        self.ctx2.stage0_file_comment = tmp_file_comment


class Context1:
    _fileTrace = None

    def __init__(self):
        self.cores = [Cores1() for i in range(constants.MAX_CORES)]
        self.n_cores = 0
        self.n_pipelines = 0
        self.pos = 0
        self.stage0_file_comment = ""
        self.stage1_file_comment = ""

        self.ctx2 = None
        self.arr_pipelines2cores = []

    def stage1_reset(self):
        for i in range(constants.MAX_CORES):
            self.cores[i].pipelines = 0
            self.cores[i].n_pipelines = 0

        self.n_cores = 0
        self.n_pipelines = 0
        self.pos = 0
        self.ctx2 = None
        # clear list
        del self.arr_pipelines2cores[:]

    def stage1_print(self):
        print('printing Context1 obj')
        print('ctx1.cores(pipelines,n_pipelines) = [ ', end='')
        for cores_count in range(0, constants.MAX_CORES):
            print('(%d,%d)' % (self.cores[cores_count].pipelines,
                               self.cores[cores_count].n_pipelines), end=' ')
        print(']')
        print('ctx1.n_cores = %d' % self.n_cores)
        print('ctx1.n_pipelines = %d' % self.n_pipelines)
        print('ctx1.pos = %d' % self.pos)
        print('ctx1.stage0_file_comment = %s' % self.stage0_file_comment)
        print('ctx1.stage1_file_comment = %s' % self.stage1_file_comment)
        if (self.ctx2 is not None):
            print('ctx1.ctx2 = ', end='')
            print(self.ctx2)
        else:
            print('ctx1.ctx2 = None')

    def stage1_init(self, c0, ctx2):
        self.stage1_reset()
        self.n_cores = 0
        while (c0.cores[self.n_cores].n_pipelines > 0):
            self.n_cores += 1

        self.n_pipelines = c0.n_pipelines
        self.ctx2 = ctx2

        self.arr_pipelines2cores = [0] * self.n_pipelines

        i = 0
        while (i < self.n_cores):
            self.cores[i].n_pipelines = c0.cores[i].n_pipelines
            i += 1

    def stage1_process(self):
        pipelines_max = len2mask(self.n_pipelines)
        while True:
            pos = 0
            overlap = 0

            if (self.cores[self.pos].pipelines == pipelines_max):
                if (self.pos == 0):
                    return

                self.cores[self.pos].pipelines = 0
                self.pos -= 1
                continue

            self.cores[self.pos].pipelines += 1
            if (popcount(self.cores[self.pos].pipelines) !=
                    self.cores[self.pos].n_pipelines):
                continue

            overlap = 0
            pos = 0
            while (pos < self.pos):
                if ((self.cores[self.pos].pipelines) &
                        (self.cores[pos].pipelines)):
                    overlap = 1
                    break
                pos += 1

            if (overlap):
                continue

            if ((self.pos > 0) and
                ((self.cores[self.pos].n_pipelines) ==
                    (self.cores[self.pos - 1].n_pipelines)) and
                    ((self.cores[self.pos].pipelines) <
                        (self.cores[self.pos - 1].pipelines))):
                continue

            if (self.pos == self.n_cores - 1):
                self.stage1_log()
                self.ctx2.stage2_init(self)
                self.ctx2.stage2_process()

                if (self.pos == 0):
                    return

                self.cores[self.pos].pipelines = 0
                self.pos -= 1
                continue

            self.pos += 1

    def stage1_log(self):
        tmp_file_comment = ""
        if(enable_stage1_traceout == 1):
            print('STAGE1: ', end='')
            tmp_file_comment += 'STAGE1: '
            i = 0
            while (i < self.n_cores):
                print('C%d = [' % i, end='')
                tmp_file_comment += "C{} = [".format(i)

                j = self.n_pipelines - 1
                while (j >= 0):
                    cond = ((self.cores[i].pipelines) & (1 << j))
                    if (cond):
                        print('1', end='')
                        tmp_file_comment += '1'
                    else:
                        print('0', end='')
                        tmp_file_comment += '0'
                    j -= 1

                print(']\t', end='')
                tmp_file_comment += ']\t'
                i += 1

            print('\n', end='')
            self.stage1_file_comment = tmp_file_comment
            self.ctx2.stage1_file_comment = tmp_file_comment

        # check if file traceing is enabled
        if(enable_stage1_fileout != 1):
            return

        # spit out the combination to file
        self.stage1_process_file()

    def stage1_updateCoresInBuf(self, nPipeline, sCore):
        rePipeline = self._fileTrace.arr_pipelines[nPipeline]
        rePipeline = rePipeline.replace("[", "\[").replace("]", "\]")
        reCore = 'core\s*=\s*((\d*)|(((s|S)\d)?(c|C)[1-9][0-9]*)).*\n'
        sSubs = 'core = ' + sCore + '\n'

        reg_pipeline = re.compile(rePipeline)
        search_match = reg_pipeline.search(self._fileTrace.in_buf)

        if(search_match):
            pos = search_match.start()
            substr1 = self._fileTrace.in_buf[:pos]
            substr2 = self._fileTrace.in_buf[pos:]
            substr2 = re.sub(reCore, sSubs, substr2, 1)
            self._fileTrace.in_buf = substr1 + substr2

    def stage1_process_file(self):
        outFileName = os.path.join(self._fileTrace.out_path,
                                   self._fileTrace.prefix_outfile)
        outFileName += "_{}CoReS".format(self.n_cores)

        i = 0  # represents core number
        while (i < self.n_cores):
            j = self.n_pipelines - 1
            pipeline_idx = 0
            while(j >= 0):
                cond = ((self.cores[i].pipelines) & (1 << j))
                if (cond):
                    # update the pipelines array to match the core
                    # only in case of cond match
                    self.arr_pipelines2cores[
                        pipeline_idx] = fileTrace.in_physical_cores[i]

                j -= 1
                pipeline_idx += 1

            i += 1

        # update the in_buf as per the arr_pipelines2cores
        for pipeline_idx in range(len(self.arr_pipelines2cores)):
            outFileName += "_{}".format(self.arr_pipelines2cores[pipeline_idx])
            self.stage1_updateCoresInBuf(
                pipeline_idx, self.arr_pipelines2cores[pipeline_idx])

        # by now the in_buf is all set to be written to file
        outFileName += self._fileTrace.suffix_outfile
        outputFile = open(outFileName, "w")

        # write out the comments
        strTruncated = ("", "(Truncated)")[self._fileTrace.ncores_truncated]
        outputFile.write(
            "; =============== Pipeline-to-Core Mapping ================\n"
            "; Generated from file {}\n"
            "; Input pipelines = {}\n"
            "; Input cores = {}\n"
            "; N_PIPELINES = {} N_CORES = {} {} hyper_thread = {}\n"
            .format(
                self._fileTrace.in_file_namepath,
                fileTrace.arr_pipelines,
                fileTrace.in_physical_cores,
                self._fileTrace.n_pipelines,
                self._fileTrace.n_cores,
                strTruncated,
                self._fileTrace.hyper_thread))

        outputFile.write(
            "; {stg0cmt}\n"
            "; {stg1cmt}\n"
            "; ========================================================\n"
            "; \n"
            .format(
                stg0cmt=self.stage0_file_comment,
                stg1cmt=self.stage1_file_comment))

        # write buffer contents
        outputFile.write(self._fileTrace.in_buf)
        outputFile.flush()
        outputFile.close()


class Context2:
    _fileTrace = None

    def __init__(self):
        self.cores = [Cores2() for i in range(constants.MAX_CORES)]
        self.n_cores = 0
        self.n_pipelines = 0
        self.pos = 0
        self.stage0_file_comment = ""
        self.stage1_file_comment = ""
        self.stage2_file_comment = ""

        # each array entry is a pipeline mapped to core stored as string
        # pipeline ranging from 1 to n, however stored in zero based array
        self.arr2_pipelines2cores = []

    def stage2_print(self):
        print('printing Context2 obj')
        print('ctx2.cores(pipelines, n_pipelines, counter, counter_max) =')
        for cores_count in range(0, constants.MAX_CORES):
            print('core[%d] = (%d,%d,%d,%d)' % (
                cores_count,
                self.cores[cores_count].pipelines,
                self.cores[cores_count].n_pipelines,
                self.cores[cores_count].counter,
                self.cores[cores_count].counter_max))

            print('ctx2.n_cores = %d' % self.n_cores, end='')
            print('ctx2.n_pipelines = %d' % self.n_pipelines, end='')
            print('ctx2.pos = %d' % self.pos)
            print('ctx2.stage0_file_comment = %s' %
                  self.self.stage0_file_comment)
            print('ctx2.stage1_file_comment = %s' %
                  self.self.stage1_file_comment)
            print('ctx2.stage2_file_comment = %s' %
                  self.self.stage2_file_comment)

    def stage2_reset(self):
        for i in range(0, constants.MAX_CORES):
            self.cores[i].pipelines = 0
            self.cores[i].n_pipelines = 0
            self.cores[i].counter = 0
            self.cores[i].counter_max = 0

            for idx in range(0, constants.MAX_PIPELINES):
                self.cores[i].bitpos[idx] = 0

        self.n_cores = 0
        self.n_pipelines = 0
        self.pos = 0
        # clear list
        del self.arr2_pipelines2cores[:]

    def bitpos_load(self, coreidx):
        i = j = 0
        while (i < self.n_pipelines):
            if ((self.cores[coreidx].pipelines) &
                    (1 << i)):
                self.cores[coreidx].bitpos[j] = i
                j += 1
            i += 1
        self.cores[coreidx].n_pipelines = j

    def bitpos_apply(self, in_buf, pos, n_pos):
        out = 0
        for i in range(0, n_pos):
            out |= (in_buf & (1 << i)) << (pos[i] - i)

        return out

    def stage2_init(self, ctx1):
        self.stage2_reset()
        self.n_cores = ctx1.n_cores
        self.n_pipelines = ctx1.n_pipelines

        self.arr2_pipelines2cores = [''] * self.n_pipelines

        core_idx = 0
        while (core_idx < self.n_cores):
            self.cores[core_idx].pipelines = ctx1.cores[core_idx].pipelines

            self.bitpos_load(core_idx)
            core_idx += 1

    def stage2_log(self):
        tmp_file_comment = ""
        if(enable_stage2_traceout == 1):
            print('STAGE2: ', end='')
            tmp_file_comment += 'STAGE2: '

            for i in range(0, self.n_cores):
                mask = len2mask(self.cores[i].n_pipelines)
                pipelines_ht0 = self.bitpos_apply(
                    (~self.cores[i].counter) & mask,
                    self.cores[i].bitpos,
                    self.cores[i].n_pipelines)

                pipelines_ht1 = self.bitpos_apply(
                    self.cores[i].counter,
                    self.cores[i].bitpos,
                    self.cores[i].n_pipelines)

                print('C%dHT0 = [' % i, end='')
                tmp_file_comment += "C{}HT0 = [".format(i)
                tmp_file_comment += bitstring_write(
                    pipelines_ht0, self.n_pipelines)

                print(']\tC%dHT1 = [' % i, end='')
                tmp_file_comment += "]\tC{}HT1 = [".format(i)
                tmp_file_comment += bitstring_write(
                    pipelines_ht1, self.n_pipelines)
                print(']\t', end='')
                tmp_file_comment += ']\t'

            print('')
            self.stage2_file_comment = tmp_file_comment

        # check if file traceing is enabled
        if(enable_stage2_fileout != 1):
            return
        # spit out the combination to file
        self.stage2_process_file()

    def stage2_updateCoresInBuf(self, nPipeline, sCore):
        rePipeline = self._fileTrace.arr_pipelines[nPipeline]
        rePipeline = rePipeline.replace("[", "\[").replace("]", "\]")
        reCore = 'core\s*=\s*((\d*)|(((s|S)\d)?(c|C)[1-9][0-9]*)).*\n'
        sSubs = 'core = ' + sCore + '\n'

        reg_pipeline = re.compile(rePipeline)
        search_match = reg_pipeline.search(self._fileTrace.in_buf)

        if(search_match):
            pos = search_match.start()
            substr1 = self._fileTrace.in_buf[:pos]
            substr2 = self._fileTrace.in_buf[pos:]
            substr2 = re.sub(reCore, sSubs, substr2, 1)
            self._fileTrace.in_buf = substr1 + substr2

    def pipelines2cores(self, n, n_bits, nCore, bHT):
        if (n_bits > 64):
            return

        i = n_bits - 1
        pipeline_idx = 0
        while (i >= 0):
            cond = (n & (1 << i))
            if (cond):
                # update the pipelines array to match the core
                # only in case of cond match
                # PIPELINE0 and core 0 are reserved
                if(bHT):
                    tmpCore = fileTrace.in_physical_cores[nCore] + 'h'
                    self.arr2_pipelines2cores[pipeline_idx] = tmpCore
                else:
                    self.arr2_pipelines2cores[pipeline_idx] = \
                        fileTrace.in_physical_cores[nCore]

            i -= 1
            pipeline_idx += 1

    def stage2_process_file(self):
        outFileName = os.path.join(self._fileTrace.out_path,
                                   self._fileTrace.prefix_outfile)
        outFileName += "_{}CoReS".format(self.n_cores)

        for i in range(0, self.n_cores):
            mask = len2mask(self.cores[i].n_pipelines)
            pipelines_ht0 = self.bitpos_apply((~self.cores[i].counter) & mask,
                                              self.cores[i].bitpos,
                                              self.cores[i].n_pipelines)

            pipelines_ht1 = self.bitpos_apply(self.cores[i].counter,
                                              self.cores[i].bitpos,
                                              self.cores[i].n_pipelines)

            # update pipelines to core mapping
            self.pipelines2cores(pipelines_ht0, self.n_pipelines, i, False)
            self.pipelines2cores(pipelines_ht1, self.n_pipelines, i, True)

        # update the in_buf as per the arr_pipelines2cores
        for pipeline_idx in range(len(self.arr2_pipelines2cores)):
            outFileName += "_{}".format(
                self.arr2_pipelines2cores[pipeline_idx])
            self.stage2_updateCoresInBuf(
                pipeline_idx, self.arr2_pipelines2cores[pipeline_idx])

        # by now the in_buf is all set to be written to file
        outFileName += self._fileTrace.suffix_outfile
        outputFile = open(outFileName, "w")

        # write the file comments
        strTruncated = ("", "(Truncated)")[self._fileTrace.ncores_truncated]
        outputFile.write(
            "; =============== Pipeline-to-Core Mapping ================\n"
            "; Generated from file {}\n"
            "; Input pipelines = {}\n"
            "; Input cores = {}\n"
            "; N_PIPELINES = {}  N_CORES = {} {} hyper_thread = {} \n"
            .format(
                self._fileTrace.in_file_namepath,
                fileTrace.arr_pipelines,
                fileTrace.in_physical_cores,
                self._fileTrace.n_pipelines,
                self._fileTrace.n_cores,
                strTruncated,
                self._fileTrace.hyper_thread))

        outputFile.write(
            "; {stg0cmt}\n"
            "; {stg1cmt}\n"
            "; {stg2cmt}\n"
            "; ========================================================\n"
            "; \n"
            .format(
                stg0cmt=self.stage0_file_comment,
                stg1cmt=self.stage1_file_comment,
                stg2cmt=self.stage2_file_comment))

        # write the buffer contents
        outputFile.write(self._fileTrace.in_buf)
        outputFile.flush()
        outputFile.close()

    def stage2_process(self):
        i = 0
        while(i < self.n_cores):
            self.cores[i].counter_max = len2mask(
                self.cores[i].n_pipelines - 1)
            i += 1

        self.pos = self.n_cores - 1
        while True:
            if (self.pos == self.n_cores - 1):
                self.stage2_log()

            if (self.cores[self.pos].counter ==
                    self.cores[self.pos].counter_max):
                if (self.pos == 0):
                    return

                self.cores[self.pos].counter = 0
                self.pos -= 1
                continue

            self.cores[self.pos].counter += 1
            if(self.pos < self.n_cores - 1):
                self.pos += 1


class FileTrace:

    def __init__(self, filenamepath):
        self.in_file_namepath = os.path.abspath(filenamepath)
        self.in_filename = os.path.basename(self.in_file_namepath)
        self.in_path = os.path.dirname(self.in_file_namepath)

        filenamesplit = self.in_filename.split('.')
        self.prefix_outfile = filenamesplit[0]
        self.suffix_outfile = ".cfg"

        # output folder:  in the same folder as input file
        # create new folder in the name of input file
        self.out_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            self.prefix_outfile)

        try:
            os.makedirs(self.out_path)
        except OSError as excep:
            if excep.errno == errno.EEXIST and os.path.isdir(self.out_path):
                pass
            else:
                raise

        self.in_buf = None
        self.arr_pipelines = []  # holds the positions of search

        self.max_cores = 15
        self.max_pipelines = 15

        self.in_physical_cores = None
        self.hyper_thread = None

        # save the num of pipelines determined from input file
        self.n_pipelines = 0
        # save the num of cores input (or the truncated value)
        self.n_cores = 0
        self.ncores_truncated = False

    def print_TraceFile(self):
        print("self.in_file_namepath = ", self.in_file_namepath)
        print("self.in_filename = ", self.in_filename)
        print("self.in_path = ", self.in_path)
        print("self.out_path = ", self.out_path)
        print("self.prefix_outfile = ", self.prefix_outfile)
        print("self.suffix_outfile = ", self.suffix_outfile)
        print("self.in_buf = ", self.in_buf)
        print("self.arr_pipelines =", self.arr_pipelines)
        print("self.in_physical_cores", self.in_physical_cores)
        print("self.hyper_thread", self.hyper_thread)


def process(n_cores, n_pipelines, fileTrace):
    '''process and map pipelines, cores.'''
    if (n_cores == 0):
        sys.exit('N_CORES is 0, exiting')

    if (n_pipelines == 0):
        sys.exit('N_PIPELINES is 0, exiting')

    if (n_cores > n_pipelines):
        print('\nToo many cores, truncating N_CORES to N_PIPELINES')
        n_cores = n_pipelines
        fileTrace.ncores_truncated = True

    fileTrace.n_pipelines = n_pipelines
    fileTrace.n_cores = n_cores

    strTruncated = ("", "(Truncated)")[fileTrace.ncores_truncated]
    print("N_PIPELINES = {}, N_CORES = {} {}"
          .format(n_pipelines, n_cores, strTruncated))
    print("---------------------------------------------------------------")

    ctx0_inst = Context0()
    ctx1_inst = Context1()
    ctx2_inst = Context2()

    # initialize the class variables
    ctx1_inst._fileTrace = fileTrace
    ctx2_inst._fileTrace = fileTrace

    ctx0_inst.stage0_init(n_cores, n_pipelines, ctx1_inst, ctx2_inst)
    ctx0_inst.stage0_process()


def validate_core(core):
    match = reg_phycore.match(core)
    if(match):
        return True
    else:
        return False


def validate_phycores(phy_cores):
    '''validate physical cores, check if unique.'''
    # eat up whitespaces
    phy_cores = phy_cores.strip().split(',')

    # check if the core list is unique
    if(len(phy_cores) != len(set(phy_cores))):
        print('list of physical cores has duplicates')
        return None

    for core in phy_cores:
        if not validate_core(core):
            print('invalid physical core specified.')
            return None
    return phy_cores


def scanconfigfile(fileTrace):
    '''scan input file for pipelines, validate then process.'''
    # open file
    filetoscan = open(fileTrace.in_file_namepath, 'r')
    fileTrace.in_buf = filetoscan.read()

    # reset iterator on open file
    filetoscan.seek(0)

    # scan input file for pipelines
    # master pipelines to be ignored
    pattern_pipeline = r'\[PIPELINE\d*\]'
    pattern_mastertype = r'type\s*=\s*MASTER'

    pending_pipeline = False
    for line in filetoscan:
        match_pipeline = re.search(pattern_pipeline, line)
        match_type = re.search('type\s*=', line)
        match_mastertype = re.search(pattern_mastertype, line)

        if(match_pipeline):
            sPipeline = line[match_pipeline.start():match_pipeline.end()]
            pending_pipeline = True
        elif(match_type):
            # found a type definition...
            if(match_mastertype is None):
                # and this is not a master pipeline...
                if(pending_pipeline):
                    # add it to the list of pipelines to be mapped
                    fileTrace.arr_pipelines.append(sPipeline)
                    pending_pipeline = False
            else:
                # and this is a master pipeline...
                # ignore the current and move on to next
                sPipeline = ""
                pending_pipeline = False
    filetoscan.close()

    # validate if pipelines are unique
    if(len(fileTrace.arr_pipelines) != len(set(fileTrace.arr_pipelines))):
        sys.exit('Error: duplicate pipelines in input file')

    num_pipelines = len(fileTrace.arr_pipelines)
    num_cores = len(fileTrace.in_physical_cores)

    print("-------------------Pipeline-to-core mapping--------------------")
    print("Input pipelines = {}\nInput cores = {}"
          .format(fileTrace.arr_pipelines, fileTrace.in_physical_cores))

    # input configuration file validations goes here
    if (num_cores > fileTrace.max_cores):
        sys.exit('Error: number of cores specified > max_cores (%d)' %
                 fileTrace.max_cores)

    if (num_pipelines > fileTrace.max_pipelines):
        sys.exit('Error: number of pipelines in input \
                cfg file > max_pipelines (%d)' % fileTrace.max_pipelines)

    # call process to generate pipeline-to-core mapping, trace and log
    process(num_cores, num_pipelines, fileTrace)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='mappipelines')

    reqNamedGrp = parser.add_argument_group('required named args')
    reqNamedGrp.add_argument(
        '-i',
        '--input-file',
        type=argparse.FileType('r'),
        help='Input config file',
        required=True)

    reqNamedGrp.add_argument(
        '-pc',
        '--physical-cores',
        type=validate_phycores,
        help='''Enter available CPU cores in
                format:\"<core>,<core>,...\"
                where each core format: \"s<SOCKETID>c<COREID>\"
                where SOCKETID={0..9}, COREID={1-99}''',
        required=True)

    # add optional arguments
    parser.add_argument(
        '-ht',
        '--hyper-thread',
        help='enable/disable hyper threading. default is ON',
        default='ON',
        choices=['ON', 'OFF'])

    parser.add_argument(
        '-nO',
        '--no-output-file',
        help='''disable output config file generation.
                Output file generation is enabled by default''',
        action="store_true")

    args = parser.parse_args()

    if(args.physical_cores is None):
        parser.error("invalid physical_cores specified")

    # create object of FileTrace and initialise
    fileTrace = FileTrace(args.input_file.name)
    fileTrace.in_physical_cores = args.physical_cores
    fileTrace.hyper_thread = args.hyper_thread

    if(fileTrace.hyper_thread == 'OFF'):
        print("!!!!disabling stage2 HT!!!!")
        enable_stage2_traceout = 0
        enable_stage2_fileout = 0
    elif(fileTrace.hyper_thread == 'ON'):
        print("!!!!HT enabled. disabling stage1 file generation.!!!!")
        enable_stage1_fileout = 0

    if(args.no_output_file is True):
        print("!!!!disabling stage1 and stage2 fileout!!!!")
        enable_stage1_fileout = 0
        enable_stage2_fileout = 0

    scanconfigfile(fileTrace)



#!/usr/bin/env python

#   BSD LICENSE
#
#   Copyright(c) 2016 Intel Corporation. All rights reserved.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     * Neither the name of Intel Corporation nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# This script creates a visual representation for a configuration file used by
# the DPDK ip_pipeline application.
#
# The input configuration file is translated to an output file in DOT syntax,
# which is then used to create the image file using graphviz
# (www.graphviz.org).
#

from __future__ import print_function
import argparse
import re
import os

#
# Command to generate the image file
#
DOT_COMMAND = 'dot -Gsize=20,30 -Tpng %s > %s'

#
# Layout of generated DOT file
#
DOT_INTRO = \
    '#\n# Command to generate image file:\n# \t%s\n#\n\n'
DOT_GRAPH_BEGIN = \
    'digraph g {\n  graph [ splines = true rankdir = "LR" ]\n'
DOT_NODE_LINK_RX = \
    '  "%s RX" [ shape = box style = filled fillcolor = yellowgreen ]\n'
DOT_NODE_LINK_TX = \
    '  "%s TX" [ shape = box style = filled fillcolor = yellowgreen ]\n'
DOT_NODE_KNI_RX = \
    '  "%s RX" [ shape = box style = filled fillcolor = orange ]\n'
DOT_NODE_KNI_TX = \
    '  "%s TX" [ shape = box style = filled fillcolor = orange ]\n'
DOT_NODE_TAP_RX = \
    '  "%s RX" [ shape = box style = filled fillcolor = gold ]\n'
DOT_NODE_TAP_TX = \
    '  "%s TX" [ shape = box style = filled fillcolor = gold ]\n'
DOT_NODE_SOURCE = \
    '  "%s" [ shape = box style = filled fillcolor = darkgreen ]\n'
DOT_NODE_SINK = \
    '  "%s" [ shape = box style = filled fillcolor = peachpuff ]\n'
DOT_NODE_PIPELINE = \
    '  "%s" [ shape = box style = filled fillcolor = royalblue ]\n'
DOT_EDGE_PKTQ = \
    '  "%s" -> "%s" [ label = "%s" color = gray ]\n'
DOT_GRAPH_END = \
    '}\n'

# Relationships between the graph nodes and the graph edges:
#
# Edge ID | Edge Label | Writer Node | Reader Node   | Dependencies
# --------+------------+-------------+---------------+--------------
# RXQx.y  | RXQx.y     | LINKx       | PIPELINEz     | LINKx
# TXQx.y  | TXQx.y     | PIPELINEz   | LINKx         | LINKx
# SWQx    | SWQx       | PIPELINEy   | PIPELINEz     | -
# TMx     | TMx        | PIPELINEy   | PIPELINEz     | LINKx
# KNIx RX | KNIx       | KNIx RX     | PIPELINEy     | KNIx, LINKx
# KNIx TX | KNIx       | PIPELINEy   | KNIx TX       | KNIx, LINKx
# TAPx RX | TAPx       | TAPx RX     | PIPELINEy     | TAPx
# TAPx TX | TAPx       | PIPELINEy   | TAPx TX       | TAPx
# SOURCEx | SOURCEx    | SOURCEx     | PIPELINEy     | SOURCEx
# SINKx   | SINKx      | PIPELINEy   | SINKx         | SINKx


#
# Parse the input configuration file to detect the graph nodes and edges
#
def process_config_file(cfgfile):
    edges = {}
    links = set()
    knis = set()
    taps = set()
    sources = set()
    sinks = set()
    pipelines = set()
    pipeline = ''

    dotfile = cfgfile + '.txt'
    imgfile = cfgfile + '.png'

    #
    # Read configuration file
    #
    lines = open(cfgfile, 'r')
    for line in lines:
        # Remove any leading and trailing white space characters
        line = line.strip()

        # Remove any comment at end of line
        line, sep, tail = line.partition(';')

        # Look for next "PIPELINE" section
        match = re.search(r'\[(PIPELINE\d+)\]', line)
        if match:
            pipeline = match.group(1)
            continue

        # Look for next "pktq_in" section entry
        match = re.search(r'pktq_in\s*=\s*(.+)', line)
        if match:
            pipelines.add(pipeline)
            for q in re.findall('\S+', match.group(1)):
                match_rxq = re.search(r'^RXQ(\d+)\.\d+$', q)
                match_swq = re.search(r'^SWQ\d+$', q)
                match_tm = re.search(r'^TM(\d+)$', q)
                match_kni = re.search(r'^KNI(\d+)$', q)
                match_tap = re.search(r'^TAP\d+$', q)
                match_source = re.search(r'^SOURCE\d+$', q)

                # Set ID for the current packet queue (graph edge)
                q_id = ''
                if match_rxq or match_swq or match_tm or match_source:
                    q_id = q
                elif match_kni or match_tap:
                    q_id = q + ' RX'
                else:
                    print('Error: Unrecognized pktq_in element "%s"' % q)
                    return

                # Add current packet queue to the set of graph edges
                if q_id not in edges:
                    edges[q_id] = {}
                if 'label' not in edges[q_id]:
                    edges[q_id]['label'] = q
                if 'readers' not in edges[q_id]:
                    edges[q_id]['readers'] = []
                if 'writers' not in edges[q_id]:
                    edges[q_id]['writers'] = []

                # Add reader for the new edge
                edges[q_id]['readers'].append(pipeline)

                # Check for RXQ
                if match_rxq:
                    link = 'LINK' + str(match_rxq.group(1))
                    edges[q_id]['writers'].append(link + ' RX')
                    links.add(link)
                    continue

                # Check for SWQ
                if match_swq:
                    continue

                # Check for TM
                if match_tm:
                    link = 'LINK' + str(match_tm.group(1))
                    links.add(link)
                    continue

                # Check for KNI
                if match_kni:
                    link = 'LINK' + str(match_kni.group(1))
                    edges[q_id]['writers'].append(q_id)
                    knis.add(q)
                    links.add(link)
                    continue

                # Check for TAP
                if match_tap:
                    edges[q_id]['writers'].append(q_id)
                    taps.add(q)
                    continue

                # Check for SOURCE
                if match_source:
                    edges[q_id]['writers'].append(q)
                    sources.add(q)
                    continue

                continue

        # Look for next "pktq_out" section entry
        match = re.search(r'pktq_out\s*=\s*(.+)', line)
        if match:
            for q in re.findall('\S+', match.group(1)):
                match_txq = re.search(r'^TXQ(\d+)\.\d+$', q)
                match_swq = re.search(r'^SWQ\d+$', q)
                match_tm = re.search(r'^TM(\d+)$', q)
                match_kni = re.search(r'^KNI(\d+)$', q)
                match_tap = re.search(r'^TAP(\d+)$', q)
                match_sink = re.search(r'^SINK(\d+)$', q)

                # Set ID for the current packet queue (graph edge)
                q_id = ''
                if match_txq or match_swq or match_tm or match_sink:
                    q_id = q
                elif match_kni or match_tap:
                    q_id = q + ' TX'
                else:
                    print('Error: Unrecognized pktq_out element "%s"' % q)
                    return

                # Add current packet queue to the set of graph edges
                if q_id not in edges:
                    edges[q_id] = {}
                if 'label' not in edges[q_id]:
                    edges[q_id]['label'] = q
                if 'readers' not in edges[q_id]:
                    edges[q_id]['readers'] = []
                if 'writers' not in edges[q_id]:
                    edges[q_id]['writers'] = []

                # Add writer for the new edge
                edges[q_id]['writers'].append(pipeline)

                # Check for TXQ
                if match_txq:
                    link = 'LINK' + str(match_txq.group(1))
                    edges[q_id]['readers'].append(link + ' TX')
                    links.add(link)
                    continue

                # Check for SWQ
                if match_swq:
                    continue

                # Check for TM
                if match_tm:
                    link = 'LINK' + str(match_tm.group(1))
                    links.add(link)
                    continue

                # Check for KNI
                if match_kni:
                    link = 'LINK' + str(match_kni.group(1))
                    edges[q_id]['readers'].append(q_id)
                    knis.add(q)
                    links.add(link)
                    continue

                # Check for TAP
                if match_tap:
                    edges[q_id]['readers'].append(q_id)
                    taps.add(q)
                    continue

                # Check for SINK
                if match_sink:
                    edges[q_id]['readers'].append(q)
                    sinks.add(q)
                    continue

                continue

    #
    # Write DOT file
    #
    print('Creating DOT file "%s" ...' % dotfile)
    dot_cmd = DOT_COMMAND % (dotfile, imgfile)
    file = open(dotfile, 'w')
    file.write(DOT_INTRO % dot_cmd)
    file.write(DOT_GRAPH_BEGIN)

    # Write the graph nodes to the DOT file
    for l in sorted(links):
        file.write(DOT_NODE_LINK_RX % l)
        file.write(DOT_NODE_LINK_TX % l)
    for k in sorted(knis):
        file.write(DOT_NODE_KNI_RX % k)
        file.write(DOT_NODE_KNI_TX % k)
    for t in sorted(taps):
        file.write(DOT_NODE_TAP_RX % t)
        file.write(DOT_NODE_TAP_TX % t)
    for s in sorted(sources):
        file.write(DOT_NODE_SOURCE % s)
    for s in sorted(sinks):
        file.write(DOT_NODE_SINK % s)
    for p in sorted(pipelines):
        file.write(DOT_NODE_PIPELINE % p)

    # Write the graph edges to the DOT file
    for q in sorted(edges.keys()):
        rw = edges[q]
        if 'writers' not in rw:
            print('Error: "%s" has no writer' % q)
            return
        if 'readers' not in rw:
            print('Error: "%s" has no reader' % q)
            return
        for w in rw['writers']:
            for r in rw['readers']:
                file.write(DOT_EDGE_PKTQ % (w, r, rw['label']))

    file.write(DOT_GRAPH_END)
    file.close()

    #
    # Execute the DOT command to create the image file
    #
    print('Creating image file "%s" ...' % imgfile)
    if os.system('which dot > /dev/null'):
        print('Error: Unable to locate "dot" executable.'
              'Please install the "graphviz" package (www.graphviz.org).')
        return

    os.system(dot_cmd)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Create diagram for IP '
                                                 'pipeline configuration '
                                                 'file.')

    parser.add_argument(
        '-f',
        '--file',
        help='input configuration file (e.g. "ip_pipeline.cfg")',
        required=True)

    args = parser.parse_args()

    process_config_file(args.file)
